#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <memory.h>

#include <windows.h>
#include <crtdbg.h>

#include <DispObj.h>

#include <mqomaterial.h>

#include <polymesh3.h>

#include <coef.h>

#define DBGH
#include <dbg.h>

#include <TexBank.h>
#include <TexElem.h>

#include <map>
#include <algorithm>
#include <iostream>
#include <iterator>

using namespace std;


extern CTexBank* g_texbank;

extern ID3DXEffect*		g_pEffect;
extern D3DXHANDLE g_hRenderL1;
extern D3DXHANDLE g_hRenderL2;
extern D3DXHANDLE g_hRenderL3;

extern D3DXHANDLE g_hdiffuse;
extern D3DXHANDLE g_hambient;
extern D3DXHANDLE g_hspecular;
extern D3DXHANDLE g_hpower;
extern D3DXHANDLE g_hemissive;
extern D3DXHANDLE g_hMeshTexture;

extern int	g_nNumActiveLights;

CDispObj::CDispObj()
{
	InitParams();
}
CDispObj::~CDispObj()
{
	DestroyObjs();
}
int CDispObj::InitParams()
{
	m_pdev = 0;
	m_pm3 = 0;//O
	
	m_dispvdecl = 0;

    m_VB = 0;
	m_IB = 0;

	return 0;
}
int CDispObj::DestroyObjs()
{
	if( m_dispvdecl ){
		m_dispvdecl->Release();
		m_dispvdecl = 0;
	}

	if( m_VB ){
		m_VB->Release();
		m_VB = 0;
	}

	if( m_IB ){
		m_IB->Release();
		m_IB = 0;
	}

	return 0;
}

int CDispObj::CreateDispObj( LPDIRECT3DDEVICE9 pdev, CPolyMesh3* pm3 )
{
	DestroyObjs();

	m_pdev = pdev;
	m_pm3 = pm3;

	CallF( CreateDecl(), return 1 );
	CallF( CreateVBandIB(), return 1 );

	return 0;
}

int CDispObj::CreateDecl()
{
	D3DVERTEXELEMENT9 vdecl[] = {
		//pos[4]
		{ 0, 0, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },

		//normal[3]
		{ 0, 16, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0 },

		//uv
		{ 0, 28, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },

		D3DDECL_END()

	};

	HRESULT hr;
	hr = m_pdev->CreateVertexDeclaration( vdecl, &m_dispvdecl );
	if( hr != D3D_OK ){
		_ASSERT( 0 );
		return 1;
	}

	return 0;
}
int CDispObj::CreateVBandIB()
{
	HRESULT hr;

	UINT elemleng;
	DWORD curFVF;

	elemleng = sizeof( PM3DISPV );
	curFVF = 0;

	hr = m_pdev->CreateVertexBuffer( m_pm3->m_optleng * elemleng,
		D3DUSAGE_WRITEONLY, curFVF,
		D3DPOOL_MANAGED,
		//D3DPOOL_DEFAULT,
		&m_VB, NULL );
	if( FAILED(hr) ){
		_ASSERT( 0 );
		return 1;
	}
	hr = m_pdev->CreateIndexBuffer( m_pm3->m_facenum * 3 * sizeof(int),
		0,
		D3DFMT_INDEX32, 
		D3DPOOL_MANAGED,
		&m_IB, NULL );
	if( FAILED(hr) ){
		_ASSERT( 0 );
		return 1;
	}

//////////
	PM3DISPV* pv;
	hr = m_VB->Lock( 0, sizeof( PM3DISPV ) * m_pm3->m_optleng, (void**)&pv, 0 );
	if( FAILED(hr) ){
		_ASSERT( 0 );
		return 1;
	}
	memcpy( pv, m_pm3->m_dispv, sizeof( PM3DISPV ) * m_pm3->m_optleng );
	m_VB->Unlock();

	int* pIndices;
	hr = m_IB->Lock( 0, m_pm3->m_facenum * 3 * sizeof(int), (void**)&pIndices, 0 );
	if( FAILED(hr) ){
		_ASSERT( 0 );
		return 1;
	}
	memcpy( pIndices, m_pm3->m_dispindex, m_pm3->m_facenum * 3 * sizeof( int ) );
	m_IB->Unlock();

	return 0;
}

int CDispObj::RenderNormal()
{
	if( !m_pm3 ){
		return 0;
	}
	if( m_pm3->m_createoptflag == 0 ){
		return 0;
	}

	HRESULT hr;
	int blno;
	for( blno = 0; blno < m_pm3->m_optmatnum; blno++ ){
		MATERIALBLOCK* currb = m_pm3->m_matblock + blno;
		CMQOMaterial* curmat = currb->mqomat;

		hr = g_pEffect->SetValue( g_hdiffuse, &(curmat->dif4f), sizeof( D3DXVECTOR4 ) );
		_ASSERT( !hr );
		hr = g_pEffect->SetValue( g_hambient, &(curmat->amb3f), sizeof( D3DXVECTOR3 ) );
		_ASSERT( !hr );
		hr = g_pEffect->SetValue( g_hspecular, &(curmat->spc3f), sizeof( D3DXVECTOR3 ) );
		_ASSERT( !hr );
		hr = g_pEffect->SetFloat( g_hpower, curmat->power );
		_ASSERT( !hr );
		hr = g_pEffect->SetValue( g_hemissive, &(curmat->emi3f), sizeof( D3DXVECTOR3 ) );
		_ASSERT( !hr );

		switch( g_nNumActiveLights ){
		case 1:
			hr = g_pEffect->SetTechnique( g_hRenderL1 );
			_ASSERT( !hr );
			break;
		case 2:
			hr = g_pEffect->SetTechnique( g_hRenderL2 );
			_ASSERT( !hr );
			break;
		case 3:
			hr = g_pEffect->SetTechnique( g_hRenderL3 );
			_ASSERT( !hr );
			break;
		default:
			_ASSERT( 0 );
			hr = g_pEffect->SetTechnique( g_hRenderL1 );
			_ASSERT( !hr );
			break;

		}

		LPDIRECT3DTEXTURE9 disptex = 0;
		if( curmat->m_texid >= 0 ){
			map<int,CTexElem*>::iterator finditr;
			finditr = g_texbank->m_texmap.find( curmat->m_texid );
			if( finditr != g_texbank->m_texmap.end() ){
				CTexElem* curte;
				curte = finditr->second;
				if( curte ){
					disptex = curte->m_ptex;
				}
			}
		}else{
			disptex = 0;
		}

		int passno;
		if( disptex ){
			passno = 0;
		}else{
			passno = 1;
		}

		hr = g_pEffect->SetTexture( g_hMeshTexture, disptex );
		_ASSERT( !hr );

		
	/////////////
		HRESULT hres;
		UINT cPasses;
		hres = g_pEffect->Begin(&cPasses, 0);
		if( hres != D3D_OK ){
			_ASSERT( 0 );
			return 1;
		}

		hres = g_pEffect->BeginPass( passno );
		if( hres != D3D_OK ){
			_ASSERT( 0 );
			return 1;
		}
		int curnumprim;
		curnumprim = currb->endface - currb->startface + 1;
		m_pdev->SetVertexDeclaration( m_dispvdecl );
		m_pdev->SetStreamSource( 0, m_VB, 0, sizeof(PM3DISPV) );
		m_pdev->SetIndices( m_IB );
		hres = m_pdev->DrawIndexedPrimitive( D3DPT_TRIANGLELIST,
											0,
											0,
											m_pm3->m_optleng,
											currb->startface * 3,
											curnumprim
											);

		hres = g_pEffect->EndPass();
		if( hres != D3D_OK ){
			_ASSERT( 0 );
			return 1;
		}

		hres = g_pEffect->End();
		if( hres != D3D_OK ){
			_ASSERT( 0 );
			return 1;
		}
	}

	return 0;
}




